# -*- coding: utf-8 -*-
"""Miscellaneous utilities used throughout the project."""

import functools
import hashlib
import urllib.request
from logging import root as logger
from typing import Union

from drytoml.paths import CACHE
from drytoml.types import Url


def cached(func):
    """Store output in drytoml's cache to use it on subsequent calls.

    Args:
        func: Function to decorate.

    Returns:
        Cached result with function result as fallback.

    .. seealso::

       * `drytoml.paths.CACHE`
       * `drytoml.app.cache`
    """

    @functools.wraps(func)
    def _wrapped(url: Url, *a, **kw):
        key = hashlib.sha256(url.encode("utf8")).hexdigest()
        path = CACHE / key
        if path.exists():
            logger.debug(
                "drytoml-cache: Using cached version of %s at %s",
                url,
                path,
            )
            with open(path) as fp:
                return fp.read()

        result = func(url, *a, **kw)
        logger.debug("Caching %s into %s", url, path)
        CACHE.mkdir(exist_ok=True, parents=True)
        with open(path, "w") as fp:
            fp.write(result)
        return result

    return _wrapped


@cached
def request(
    url: Union[str, Url],
) -> str:
    """Request a `url` using a GET.

    Args:
        url: The URL to GET.

    Returns:
        Decoded content.
    """

    request_ = urllib.request.Request(Url(url))
    # avoid server-side caching
    request_.add_header("Pragma", "no-cache")
    request_.add_header("User-Agent", "Mozilla/5.0")
    with urllib.request.urlopen(request_) as response:  # noqa: S310
        contents = response.read().decode("utf-8")
    return contents



"""This module contains the `export` command and its required utilities."""
import logging

from drytoml.parser import DEFAULT_EXTEND_KEY
from drytoml.parser import Parser


def export(
    file="pyproject.toml",
    key=DEFAULT_EXTEND_KEY,
) -> str:
    """Generate resulting TOML after transclusion.

    Args:
        file: TOML file to transclude values.
        key: Name too look for inside the file to activate interpolation.

    Returns:
        The transcluded toml.

    Example:
        >>> toml = export("isort.toml", "base")
    """

    logging.basicConfig(level=60, format="%(message)s", force=True)
    return Parser.from_file(file, extend_key=key).parse().as_string()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Cli application for drytoml."""
import argparse
import logging
import sys
from typing import List
from typing import Optional

import fire

from drytoml import logger
from drytoml.app.cache import Cache
from drytoml.app.explain import explain
from drytoml.app.export import export
from drytoml.app.wrappers import black
from drytoml.app.wrappers import check
from drytoml.app.wrappers import flake8helled
from drytoml.app.wrappers import flakehell
from drytoml.app.wrappers import isort
from drytoml.app.wrappers import pylint

INTERNAL_CMDS = {
    cmd.__name__.lower(): cmd
    for cmd in (
        Cache,
        explain,
        export,
        check,
    )
}

WRAPPERS = {
    cmd.__name__.lower(): cmd
    for cmd in (
        black,
        isort,
        pylint,
        flakehell,
        flake8helled,
    )
}


def setup_log(argv: Optional[List[str]]) -> List[str]:
    """Control verbosity via logging level using "-q/-v" as flags.

    Args:
        argv: If not set, use sys.argv. For each "-v" or "--verbose",
              increase the log level verbosity. If it contains a "-q",
              or a "--quiet", set lefel to `logging.CRITICAL`.

    Returns:
        Unparsed, remanining arguments.
    """

    argv = argv or sys.argv

    parser = argparse.ArgumentParser(add_help=False)
    parser.add_mutually_exclusive_group()
    parser.add_argument("-v", "--verbose", action="count", default=0)
    parser.add_argument("-q", "--quiet", action="count", default=0)
    args, unknown = parser.parse_known_args()
    level = max(1, 20 + 50 * args.quiet - 10 * args.verbose)
    logging.basicConfig(level=level, format="%(message)s", force=True)
    logger.debug(
        "drytoml: Log level set to %s because %s", level, args.__dict__
    )
    return sys.argv[:1] + unknown


def main():
    """Execute the cli application.

    Returns:
        The result of the wrapped command
    """
    sys.argv = setup_log(sys.argv)

    if len(sys.argv) == 1 or sys.argv[1] not in WRAPPERS:
        return fire.Fire(INTERNAL_CMDS)

    del sys.argv[0]
    return WRAPPERS[sys.argv[0]]()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""Utilities and logic for handling inter-toml merges."""

from typing import Dict
from typing import List
from typing import Union

from tomlkit.container import Container
from tomlkit.container import OutOfOrderTableProxy
from tomlkit.items import AoT
from tomlkit.items import Array
from tomlkit.items import Bool
from tomlkit.items import Date
from tomlkit.items import DateTime
from tomlkit.items import Float
from tomlkit.items import Integer
from tomlkit.items import Item
from tomlkit.items import Key
from tomlkit.items import Null
from tomlkit.items import String
from tomlkit.items import Table
from tomlkit.items import Time
from tomlkit.toml_document import TOMLDocument

from drytoml.locate import deep_del

RAW_ITEMS = (
    Integer,
    Float,
    Bool,
    DateTime,
    Date,
    Time,
    String,
    Null,
)


def deep_merge(current: Item, incoming: Item) -> Item:
    """Merge two items using a type-dependent strategy.

    Args:
        current: Item to merge into.
        incoming: Item to merge from.

    Raises:
        NotImplementedError: Unable to merge received current and
            incoming item given their types.

    Returns:
        The current Item, after merging in-place.

    """

    if isinstance(current, list):
        if isinstance(incoming, list):
            return deep_extend(current, incoming)

    if isinstance(current, (Table, TOMLDocument, OutOfOrderTableProxy)):
        if isinstance(incoming, (Table, TOMLDocument, OutOfOrderTableProxy)):
            for key in {*current.keys(), *incoming.keys()}:
                if key not in incoming:
                    continue
                if key not in current:
                    # emulate incoming container skeleton
                    current.append(key, incoming[key])
                    continue
                current[key] = deep_merge(current[key], incoming[key])
            return current

    if isinstance(current, RAW_ITEMS):
        if isinstance(incoming, RAW_ITEMS):
            return current

    raise NotImplementedError


def merge_targeted(
    document: Container,
    incoming: Container,
    breadcrumbs: List[Union[str, int]],
) -> TOMLDocument:
    """Merge specific path contents from an incoming contianer into another.

    Args:
        document: The container to store the merge result.
        incoming: The source of the incoming data.
        breadcrumbs: Location of the incoming contend.

    Returns:
        The `document`, after merging in-place.
    """

    if not breadcrumbs:
        return deep_merge(document, incoming)

    location = document
    incoming_data = incoming
    crumbs, final = breadcrumbs[:-1], breadcrumbs[-1]

    for key in crumbs:
        incoming_data = incoming_data[key]
        if key not in location:
            # emulate incoming container skeleton
            location[key] = type(incoming_data)()
        location = location[key]

    if final not in location:
        location[final] = incoming_data[final]
    else:
        location[final] = deep_merge(location[final], incoming_data[final])

    return document


def deep_extend(
    current: Union[Array, AoT],
    incoming: Union[Array, AoT],
) -> Union[Array, AoT]:
    """Extend a container with another's contents.

    Args:
        current: Container to extend (in-place).
        incoming: Container to extend with.

    Returns:
        The recevied container, modified in-place.

    """
    current.extend(incoming)
    return current


class TomlMerger:
    """Encapsulate toml merging strategies and procedures."""

    def merge_simple(
        self,
        value: Item,
        breadcrumbs: List[Key],
    ):
        """Merge a value in a specific position of the container.

        Args:
            value: Incoming data to be merged.
            breadcrumbs: Location of the parent container for the
                incoming value merge.
        """
        incoming_parser = self.build_subparser(value)
        incoming = incoming_parser.parse()
        merge_targeted(self.container, incoming, breadcrumbs)

    def merge_list_like(
        self,
        values: List[Item],
        breadcrumbs: List[Key],
    ):
        """Merge sequence of values in a specific position of the container.

        Args:
            values: Incoming data to be merged one by one, in reversed order.
            breadcrumbs: Location of the parent container for the
                    incoming value merge.
        """
        for val in reversed(values):
            self.merge_simple(
                val,
                breadcrumbs,
            )

    def merge_dict_like(
        self,
        dct: Dict[Key, Item],
        breadcrumbs: List[Key],
    ):
        """Merge a dict-like object into a specific container position.

        Args:
            dct: The incoming data.
            breadcrumbs: Location of the parent container for the
                    incoming data merge.
        """
        for key, val in dct.items():
            self(
                val,
                [*breadcrumbs, key],
            )

    def __init__(
        self,
        container: Container,
        parser,
    ):
        """Construct a merger which can transclude a TOML.

        Args:
            container: The outermost data container.
            parser: The parsed to be used for child parsed when
                transcluding toml objects.

        """
        self.container = container
        self.parser = parser

        strategies = {
            self.merge_simple: RAW_ITEMS,
            self.merge_list_like: (AoT, Array),
            self.merge_dict_like: (dict,),
        }
        self.merge_options = {
            item_type: solver
            for solver, item_types in strategies.items()
            for item_type in item_types
        }

    def build_subparser(self, value: Item):
        """Construct child parser from specific content.

        Args:
            value:The content of the TOML data to be parsed.

        Returns:
            The instantiated child parser.
        """

        return type(self.parser).factory(
            value,
            self.parser.extend_key,
            self.parser.reference,
            level=self.parser.level + 1,
        )

    def __call__(
        self,
        value: Item,
        breadcrumbs: List[Key],
        delete_dangling: bool = False,
    ):
        """Run the merging strategies.

        Args:
            value: Incoming value to merge into `self.container`.
            breadcrumbs: Location of the `base_key` which triggers the
                merge.
            delete_dangling: If set, removes the predefined base key
                which triggered the transclusions after merge is
                complete.

        Raises:
            NotImplementedError: Unable to merge given value type.
        """
        extend_key = self.parser.extend_key

        incoming_class = type(value)
        try:
            merge = self.merge_options[incoming_class]
        except KeyError as exc:
            raise NotImplementedError(
                "Unable to merge {} into {}".format(
                    incoming_class, type(self.container)
                )
            ) from exc

        merge(value, breadcrumbs)

        if delete_dangling:
            deep_del(self.container, extend_key, *breadcrumbs)



# -*- coding: utf-8 -*-
"""Manage drytoml's internal cache.

This module contains the Cache class, which allows fire to execute any
method (bound, static, or classmethod) as sub-command from the cli.
"""

import shutil
import sys
from pathlib import Path
from typing import Dict
from typing import Union

from drytoml import logger
from drytoml.paths import CACHE


class Cache:
    """Manage drytoml's internal cache."""

    @classmethod
    def clear(
        cls, force: bool = False, name: str = ""
    ) -> Dict[Union[Path, str], str]:
        """Clear drytoml's cache.

        Args:
            force: Clear without asking.
            name: If set, only clear a specific cache element.

        Returns:
            Contents of the cache after clearing it.

        """

        if not force:
            truthy = {"t", "true", "y", "yes", "clear"}
            response = input("Clear cache? [y/N]")
            if response.lower() not in truthy:
                logger.error("Aborted")
                sys.exit(1)

        worked = False
        for descendant in CACHE.glob("**/*"):
            if name and (descendant.stem != Path(name).stem):
                continue

            if descendant.is_file():
                descendant.unlink()
                worked = True
            elif descendant.is_dir():
                shutil.rmtree(str(descendant.resolve()))
                worked = True
        if worked:
            logger.info("Succesfully cleared %s %s", CACHE, name)
        else:
            logger.info("Nothing cleared from %s", CACHE)
            sys.exit(1)

        return cls.show()

    @staticmethod
    def show() -> Dict[Union[str, Path], str]:
        """Show drytoml's cache contents.

        Returns:
            Locations -> weight (in kb) mapping
        """
        data = {
            descendant: descendant.stat().st_size
            for descendant in CACHE.glob("**/*")
            if descendant.is_file()
        }
        if not data:
            return logger.info("Cache is empty: %s", CACHE)
        info = {
            **{k: f"{v/1024:.2f} kb" for k, v in data.items()},
            "__total__": f"{sum(data.values()) / 1024:.2f} kb",
        }
        return info



# -*- coding: utf-8 -*-
"""Additional Source to transclude tomlkit with URL and files."""

from pathlib import Path
from textwrap import dedent as _
from typing import Optional
from typing import Union

from tomlkit.parser import Parser as BaseParser
from tomlkit.toml_document import TOMLDocument

from drytoml import logger
from drytoml.locate import deep_find
from drytoml.merge import TomlMerger
from drytoml.types import Url
from drytoml.utils import request

DEFAULT_EXTEND_KEY = "__extends"


class Parser(BaseParser):
    """Extend tomlkit parser to allow transclusion."""

    def __init__(
        self,
        string: str,
        extend_key=DEFAULT_EXTEND_KEY,
        reference: Optional[Union[str, Path, Url]] = None,
        level=0,
    ):
        """Construct a transclusion-enabled toml parser.

        Args:
            string: Raw toml content.
            extend_key: key to look for to init transclusion.
            reference: Reference for the source of the content
                (eg url, file, etc).
            level: Number of parent documents previously parsed to
                instantiate this.
        """
        self.extend_key = extend_key
        self.reference = reference or Path.cwd()
        self.from_string = not reference
        self.level = level
        super().__init__(string)

    def __repr__(self) -> str:
        """Enable parser visual differentiation from repr.

        Returns:
            A string of the form ``Parser(reference, extend_key)``

        Examples:

        >>> parser = Parser.from_file("pyproject.toml")
        >>> parser
        Parser('pyproject.toml', extend_key='__extends')

        >>> content = parser.parse().as_string()
        >>> Parser(content)
        Parser('/path/to/dir' as cwd, (from string), extend_key='__extends')

        >>> Parser.factory("https://github.com/pytest-dev/pytest/blob/master/pyproject.toml")
        Parser('https://github.com/pytest-dev/pytest/blob/master/pyproject.toml', extend_key='__extends')
        """
        return "{}Parser('{}'{}, extend_key='{}')".format(
            self._log_indent,
            self.reference,
            " as cwd, (from string)" if self.from_string else "",
            self.extend_key,
        )

    @classmethod
    def from_file(cls, path, extend_key=DEFAULT_EXTEND_KEY, level=0):
        """Instantiate a parser from file.

        Args:
            path: Path to an existing file with the toml contents.
            extend_key: kwarg to construct the parser.
            level: kwarg to construct the parser.

        Returns:
            Parser instantiated from received path.

        """
        with open(path) as fp:
            raw = fp.read()
        return cls(raw, extend_key=extend_key, reference=path, level=level)

    @classmethod
    def from_url(cls, url, extend_key=DEFAULT_EXTEND_KEY, level=0):
        """Instantiate a parser from url.

        Args:
            url: URL to an existing file with the toml contents.
            extend_key: kwarg to construct the parser.
            level: kwarg to construct the parser.

        Returns:
            Parser instantiated from received url.
        """
        raw = request(url)
        return cls(raw, extend_key=extend_key, reference=url, level=level)

    @classmethod
    def factory(
        cls,
        reference: Union[str, Url, Path],
        extend_key=DEFAULT_EXTEND_KEY,
        parent_reference: Optional[Union[str, Path, Url]] = None,
        level=0,
    ):
        """Instantiate a parser from url, string, or path.

        Args:
            reference: Existing file/url/path with the toml contents.
            extend_key: kwarg to construct the parser.
            parent_reference: Used to parse relative paths.
            level: kwarg to construct the parser.

        Returns:
            Parser instantiated from received reference.

        Raises:
            ValueError: Attempted to intantiate a parser with a relative
                path as reference, without a parent reference.
        """

        if Url.validate(reference):
            return cls.from_url(
                reference,
                extend_key=extend_key,
                level=level,
            )

        path = Path(reference)
        if not path.is_absolute():
            if not parent_reference:
                raise ValueError("Must supply absolute path or parent")
            path = (parent_reference.parent / path).resolve()

        return cls.from_file(path, extend_key=extend_key, level=level)

    @property
    def _log_indent(self):
        return " " * 2 * self.level

    def _log_document(self, document):
        raw = document.as_string()
        return _(
            f"""
{"="*30}{self} CONTENTS STARTS HERE{"="*30}
{raw}
{"="*30}{self} CONTENTS END HERE{"="*30}"""
        ).replace("\n", f"\n{self._log_indent}")

    def parse(self) -> TOMLDocument:
        """Parse recursively until no transclusions are required.

        Returns:
            The parsed, transcluded document.
        """
        document = super().parse()
        logger.info("%s: Parsing started", self)
        logger.debug(
            "%s: Source contents:\n\n%s", self, self._log_document(document)
        )

        while True:
            base_key_locations = sorted(
                deep_find(document, self.extend_key),
                key=lambda path_ct: path_ct[0],
            )

            if not base_key_locations:
                logger.debug("%s: No %s found", self, self.extend_key)
                break
            logger.info(
                "%s: Found '%s': at %s",
                self,
                self.extend_key,
                [
                    ".".join(crumbs_val[0]) or "(document root)"
                    for crumbs_val in base_key_locations
                ],
            )

            for breadcrumbs, value in base_key_locations:
                logger.debug(
                    "%s: Before merging %s contents:\n\n%s",
                    self,
                    breadcrumbs,
                    self._log_document(document),
                )
                merge = TomlMerger(document, self)
                merge(value, breadcrumbs, delete_dangling=True)
                logger.debug(
                    "%s: After merging %s contents:\n\n%s",
                    self,
                    breadcrumbs,
                    self._log_document(document),
                )

        logger.info("%s: Parsing finished", self)
        logger.debug(
            "%s: Final contents:\n\n%s",
            self,
            self._log_document(document),
        )
        return document



# -*- coding: utf-8 -*-
"""Custom types and synonyms."""
import re


class Url(str):
    """Avoid instantiation for non-compliant url strings."""

    URL_VALIDATOR = re.compile(
        r"^(?:http|ftp)s?://"  # http:// or https://
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"  # noqa: C0301  # domain...
        r"localhost|"  # localhost...
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
        r"(?::\d+)?"  # optional port
        r"(?:/?|[/?]\S+)$",
        re.IGNORECASE,
    )
    """Django validator."""

    def __init__(self, string):
        """Validate string as url before instantiating.

        Args:
            string: Url to validate

        Raises:
            ValueError: The received string is not a valid URL.
        """
        if not self.validate(string):
            raise ValueError("Not a valid URL")
        super().__init__()

    @classmethod
    def validate(
        cls,
        maybe_url,
    ) -> bool:
        """Validate url string using django regex.

        Args:
            maybe_url: Url to validate.

        Returns:
            `True` iff validation succeeds.
        """
        return cls.URL_VALIDATOR.match(str(maybe_url)) is not None



# -*- coding: utf-8 -*-
"""Utilities to simplify deep `getitem` calls."""

from typing import Generator
from typing import List
from typing import Optional
from typing import Union

from tomlkit.items import Key


def deep_find(
    container: Union[
        str,
        list,
        dict,
    ],
    extend_key: str,
    breadcrumbs: Optional[List[str]] = None,
) -> Generator[List[Key], type, None]:
    """Walk a data structure to yield all occurences of a key.

    Args:
        container: Where to look for the key.
        extend_key: The key to look for.
        breadcrumbs: The sequence of walked keys to the current position.

    Returns:
        The container type

    Yields:
        Tuple containing (`breadcrumbs`, `content`) where `extend_key`
            was found.

    """
    breadcrumbs = breadcrumbs or []

    if isinstance(
        container,
        list,
    ):
        for index, element in enumerate(container):
            yield from deep_find(element, extend_key, [*breadcrumbs, index])
        return list

    if isinstance(
        container,
        dict,
    ):
        for name, content in container.items():
            yield from deep_find(content, extend_key, [*breadcrumbs, name])

            if name == extend_key:
                yield breadcrumbs, content

        return dict

    return type(container)


def deep_del(document, final, *breadcrumbs):
    """Delete content located deep within a data structure.

    Args:
        document: Where to remove the content from.
        final: Last key required to locate the element to be deleted.
        breadcrumbs: The path to walk from the container root up to the
            parent ob the object to be deleted.

    Examples:
        Examples should be written in doctest format, and should illustrate how
        to use the function.

        >>> container = {
        ...     "foo": [
        ...         {},
        ...         {"bar": "delete_me"}
        ...     ]
        ... }
        {'foo': [{}, {'bar': 'delete_me'}]}
        >>> deep_del(container, "bar", ["foo", 1])
        >>> container
        {'foo': [{}, {}]}
    """
    current = document
    for key in breadcrumbs:
        current = current[key]
    del current[final]



"""Call drytoml's entrypoint as a python module.

Example:

    ```console
    $ python -m dry
    ```
"""
import sys

from drytoml.app import main

if __name__ == "__main__":
    sys.argv[0] = "dry"
    main()



"""Common filesystem paths for drytoml."""

import os
from pathlib import Path
from typing import Union


def env_or(xdg_env: str, home_subdir: Union[str, Path]) -> Path:
    """Retreive path from xdg env, with home subdir as default.

    Args:
        xdg_env: Name of the environment variable.
        home_subdir: Sub-directory (inside $HOME) to use as default
            value if the env var is not found.

    Returns:
        Resulting path

    Examples:

        For an exising env var:
        >>> env_or("XDG_CACHE_HOME", ".cache/custom")
        PosixPath('/home/vscode/.cache')

        For an env var not present:
        >>> env_or("XDG_DATA_HOME", ".custom_subdir")
        PosixPath('/home/vscode/.custom_subdir')
    """
    return Path(os.environ.get(xdg_env, Path.home() / home_subdir))


CACHE = env_or("XDG_CACHE_HOME", ".cache") / "drytoml"
"""Location of drytoml's cache.
It can be overriden by changing the XDG_CACHE_HOME env var.
"""

CONFIG = env_or("XDG_CONFIG_HOME", ".config") / "drytoml"
"""Location of drytoml's configuration files.
It can be overriden by changing the XDG_CONFIG_HOME env var.
"""



"""This module contains the `explain` command and its required utilities."""
from drytoml.parser import DEFAULT_EXTEND_KEY
from drytoml.parser import Parser


def explain(
    file="pyproject.toml",
    key=DEFAULT_EXTEND_KEY,
):
    """Show steps for toml transclusion.

    Args:
        file: TOML file to interpolate values.
        key: Name too look for inside the file to activate interpolation.

    Example:
        >>> explain("isort.toml", "base")
    """
    Parser.from_file(file, extend_key=key).parse()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['drytoml', 'drytoml.app']

package_data = \
{'': ['*']}

install_requires = \
['fire>=0.4.0,<0.5.0', 'tomlkit>=0.7.0,<0.8.0']

extras_require = \
{'all': ['black>=20.8b1,<21.0',
         'flakeheaven>=0.10.0-alpha.0,<0.11.0',
         'isort>=5.7.0,<6.0.0'],
 'black': ['black>=20.8b1,<21.0'],
 'dev': ['black>=20.8b1,<21.0',
         'flakeheaven>=0.10.0-alpha.0,<0.11.0',
         'isort>=5.7.0,<6.0.0',
         'pytest>=6.2.2,<7.0.0',
         'Sphinx>=3.4.3,<4.0.0',
         'pytest-cov>=2.11.1,<3.0.0',
         'sphinx-rtd-theme>=0.5.1,<0.6.0',
         'flake8-bandit>=2.1.2,<3.0.0',
         'flake8-bugbear>=20.11.1,<21.0.0',
         'flake8-builtins>=1.5.3,<2.0.0',
         'flake8-comprehensions>=3.3.1,<4.0.0',
         'darglint>=1.6.0,<2.0.0',
         'flake8-docstrings>=1.5.0,<2.0.0',
         'flake8-eradicate>=1.0.0,<2.0.0',
         'flake8-mutable>=1.2.0,<2.0.0',
         'flake8-debugger>=4.0.0,<5.0.0',
         'flake8-pytest-style>=1.3.0,<2.0.0',
         'pep8-naming>=0.11.1,<0.12.0',
         'pytest-html>=3.1.1,<4.0.0',
         'm2r2>=0.2.7,<0.3.0',
         'recommonmark>=0.7.1,<0.8.0',
         'commitizen>=2.14.2,<3.0.0',
         'pre-commit>=2.10.1,<3.0.0',
         'pylint>=2.6.0,<3.0.0'],
 'docs': ['Sphinx>=3.4.3,<4.0.0',
          'sphinx-rtd-theme>=0.5.1,<0.6.0',
          'm2r2>=0.2.7,<0.3.0',
          'recommonmark>=0.7.1,<0.8.0'],
 'flakehell': ['flakeheaven>=0.10.0-alpha.0,<0.11.0'],
 'format': ['black>=20.8b1,<21.0', 'isort>=5.7.0,<6.0.0'],
 'isort': ['isort>=5.7.0,<6.0.0'],
 'lint': ['black>=20.8b1,<21.0',
          'flakeheaven>=0.10.0-alpha.0,<0.11.0',
          'isort>=5.7.0,<6.0.0',
          'pytest>=6.2.2,<7.0.0',
          'flake8-bandit>=2.1.2,<3.0.0',
          'flake8-bugbear>=20.11.1,<21.0.0',
          'flake8-builtins>=1.5.3,<2.0.0',
          'flake8-comprehensions>=3.3.1,<4.0.0',
          'darglint>=1.6.0,<2.0.0',
          'flake8-docstrings>=1.5.0,<2.0.0',
          'flake8-eradicate>=1.0.0,<2.0.0',
          'flake8-mutable>=1.2.0,<2.0.0',
          'flake8-debugger>=4.0.0,<5.0.0',
          'flake8-pytest-style>=1.3.0,<2.0.0',
          'pep8-naming>=0.11.1,<0.12.0',
          'pylint>=2.6.0,<3.0.0'],
 'test': ['pytest>=6.2.2,<7.0.0',
          'pytest-cov>=2.11.1,<3.0.0',
          'pytest-html>=3.1.1,<4.0.0']}

entry_points = \
{'console_scripts': ['dry = drytoml.app:main']}

setup_kwargs = {
    'name': 'drytoml',
    'version': '0.2.6',
    'description': 'Keep toml configuration centralized and personalizable',
    'long_description': '# drytoml\n\nKeep your toml configuration centralized and personalizable.\n\n---\n\n[![PyPI](https://img.shields.io/pypi/v/drytoml?color=yellow)](https://pypi.org/project/drytoml/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/drytoml)](https://www.python.org/downloads/)\n\n---\n\n[![ReadTheDocs](https://readthedocs.org/projects/drytoml/badge/?version=latest)](https://drytoml.readthedocs.io/en/latest/)\n\n---\n\n[![Format](https://github.com/pwoolvett/drytoml/workflows/Format/badge.svg)](https://github.com/pwoolvett/drytoml/actions?query=workflow%3AFormat)\n[![Lint](https://github.com/pwoolvett/drytoml/workflows/Lint/badge.svg)](https://github.com/pwoolvett/drytoml/actions?query=workflow%3ALint)\n[![Test](https://github.com/pwoolvett/drytoml/workflows/Test/badge.svg)](https://github.com/pwoolvett/drytoml/actions?query=workflow%3ATest)\n[![codecov](https://codecov.io/gh/pwoolvett/drytoml/branch/master/graph/badge.svg?token=YXO8NDUU0G)](https://codecov.io/gh/pwoolvett/drytoml)\n\n---\n\n[![VSCode Ready-to-Code](https://img.shields.io/badge/devcontainer-enabled-blue?logo=docker)](https://code.visualstudio.com/docs/remote/containers)\n[![License: Unlicense](https://img.shields.io/badge/license-UNLICENSE-white.svg)](http://unlicense.org/)\n\n\n---\n\n\nThrough `drytoml`, TOML files can have references to any content from another TOML file.\nReferences work with relative/absolute paths and urls, and can point to whole files, a\nspecific table, or in general anything reachable by a sequence of `getitem` calls, like\n`["tool", "poetry", "source", 0, "url"]`. `drytoml` takes care of transcluding the\ncontent for you.\n\nInspired by [flakehell](https://pypi.org/project/flakehell/) and\n[nitpick](https://pypi.org/project/nitpick/), the main motivation behind `drytoml` is to\nhave several projects sharing a common, centralized configuration defining codestyles,\nbut still allowing granular control when required.\n\nThis is a deliberate departure from [nitpick](https://pypi.org/project/nitpick/), which\nworks as a linter instead, ensuring your local files have the right content.\n\n\n## Usage\n\n`drytoml` has two main usages:\n\n1. Use a file as a reference and "build" the resulting one:\n\n    ```toml\n    # contents of pyproject.dry.toml\n    ...\n    [tool.black]\n    __extends = "../../common/black.toml"\n    target-version = [\'py37\']\n    include = \'\\.pyi?$\'\n    ...\n    ```\n\n    ```toml\n    # contents of ../../common/black.toml\n    [tool.black]\n    line-length = 100\n    ```\n\n   ```console\n   $ dry export --file=pyproject.dry.toml > pyproject.toml\n   ```\n\n    ```toml\n    # contents of pyproject.toml\n    [tool.black]\n    line-length = 100\n    target-version = [\'py37\']\n    include = \'\\.pyi?$\'\n    ```\n\n2. Use included wrappers, allowing you to use references in your current configuration\n   without changing any file:\n\n   Consider the original `pyproject.dry.toml` from the example above, an alternative\n   usage for `drytoml` is shown next. Instead of this:\n\n   ```console\n   $ black .\n   All done! ✨ 🍰 ✨\n   14 files left unchanged.\n   ```\n\n   You would run this:\n\n   ```console\n   $ dry black\n   reformatted /path/to/cwd/file_with_potentially_long_lines.py\n   reformatted /path/to/cwd/another_file_with_potentially_long_lines.py\n   All done! ✨ 🍰 ✨\n   2 files reformatted, 12 files left unchanged.\n   ```\n\n   What just happened? `drytoml` comes with a set of wrappers which\n\n   1. Create a transcluded temporary file, equivalent to the resulting `pyproject.toml`\n      in the example above\n   2. Configure the wrapped tool (`black` in this case) to use the temporary file\n   3. Run `black`, and get rid of the file on exit.\n\n\nFor the moment, the following wrappers are available (more to come, contributions are\nwelcome):\n\n* [x] [black](https://github.com/psf/black)\n* [x] [isort](https://pycqa.github.io/isort/)\n* [x] [flakehell, flake8helled](https://github.com/life4/flakehell) *\n\nIn the works:\n* [ ] docformatter\n* [ ] pytest\n\n### Notes\n\nAlthough the flakehell project was archived, we\'re keeping a fork alive from\n[here](https://github.com/pwoolvett/flakehell), availabe as\n[flakeheaven](https://pypi.org/project/flakeheaven) in pypi.\n\n\n## Setup\n\n### Prerequisites\n\n  * A compatible python >3.6.9\n  * [recommended] virtualenv\n  * A recent `pip`\n\n### Install\n\n  Install as usual, with `pip`, `poetry`, etc:\n\n* `pip install drytoml`\n* `poetry add drytoml` (probably you\'ll want `poetry add --dev drytoml` instead)\n\n## Usage\n\nFor any command , run `--help` to find out flags and usage.\nSome of the most common are listed below:\n\n* Use any of the provided wrappers as a subcommand, eg `dry black` instead of `black`.\n* Use `dry -q export` and redirect to a file, to generate a new file with transcluded\n  contents\n* Use `dry cache` to manage the cache for remote references.\n\n\n\n## FAQ\n\n**Q: I want to use a different key to trigger transclusions**\n\n   A: In cli mode (using the `dry` command), you can pass the `--key` flagcli, to change\n   it. In api mode (from python code), initialize `drytoml.parser.Parser` using a\n   custom value for the `extend_key` kwarg.\n\n\n**Q: I changed a referenced toml upstream (eg in github) but still get the same result.**\n\n   A: Run `dry cache clear --help` to see available options.\n\n## Contribute\n\nStart by creating an issue, forking the project and creating a Pull Request.\n\n### Setting up the development environment\n\nIf you have docker, the easiest way is to use the provided devcontainer inside vscode,\nwhich already contains everything pre-installed. You must open the cloned directory\nusing the [remote-containers extension](https://marketplace.visualstudio.com/items?itemName=ms-vscode-remote.remote-containers).\nJust run `poetry shell` or prepend any command with `poetry run` to ensure commands\nare run inside the virtual environment.\n\nAlternatively, you can use poetry: `poetry install -E dev`\n\nThe next steps assume you have already activated the venv.\n\n### Committing\n\n* Commits in every branch must adhere to [Conventional Commits](https://www.conventionalcommits.org/en/v1.0.0/).\n  Releases are done automatically and require conventional commits messages compliance.\n\n* To validate commits, you can install the pre-commit hook\n\n  ```console\n  pre-commit install --hook-type commit-msg\n  ```\n\n* With venv activated, you can commit using `cz commit` instead of `git commit` to\n  ensure compliance.\n\n### Running checks\n\nYou can look at the different actions defined in `.github/workflows`. There are three\nways to check your code:\n\n* Remotely, by pushing to an open PR. You\'ll se the results in the PR page.\n\n* Manually, executing the check from inside a venv\n\n  For example, to generate the documentation, check `.github/workflows/docs`. To run the\n  `Build with Sphinx` step:\n\n  ```console\n  sphinx-build docs/src docs/build\n  ```\n\n  Or to run pytest, from `.github/workflows/tests.yml`:\n\n  ```console\n  sphinx-build docs/src docs/build\n  ```\n\n  ... etc\n\n* Locally, with [act](https://github.com/nektos/act) (Already installed in the\n  devcontainer)\n\nFor example, to emulate a PR run for the tests workflow:\n  \n ```console\n act -W .github/workflows/tests.yml pull_request\n ```\n\n## TODO\n\nCheck out current development [here](https://github.com/pwoolvett/drytoml/projects/2)\n',
    'author': 'Pablo Woolvett',
    'author_email': 'pablowoolvett@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pwoolvett/drytoml',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.9,<4.0.0',
}


setup(**setup_kwargs)



"""Third-party commands enabled through drytoml."""

import importlib
import os
import shlex
import subprocess as sp  # noqa: S404
import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import IO
from typing import Callable
from typing import List
from typing import Union

from drytoml.parser import Parser


def import_callable(string: str) -> Callable:
    """Import a module from a string using colon syntax.

    Args:
        string: String of the form `package.module:object`

    Returns:
        The imported module

    """
    module_str, tool_main_str = string.split(":")
    module = importlib.import_module(module_str)
    tool_main = getattr(module, tool_main_str)
    return tool_main


class Wrapper:
    """Common skeleton for third-party wrapper commands."""

    cfg: str
    virtual: IO[str]

    def __call__(self, importstr):
        """Execute the wrapped callback.

        Args:
            importstr: String of the form `package.module:object`

        .. seealso:: `import_callable`

        """

        with self.tmp_dump() as virtual:
            self.virtual = virtual
            self.pre_import()
            self.pre_call()
            tool_main = import_callable(importstr)
            sys.exit(tool_main())

    def pre_import(self):
        """Execute custom processing done before callback import."""

    def pre_call(self):
        """Execute custom processing done before callback execut."""

    @contextmanager
    def tmp_dump(self):
        """Yield a temporary file with the configuration toml contents.

        Yields:
            Temporary file with the configuration toml contents
        """
        parser = Parser.from_file(self.cfg)
        document = parser.parse()

        # ensure locally referenced files work
        path = Path(self.cfg)
        if path.is_absolute():
            parent = path.parent
        else:
            parent = (Path.cwd() / self.cfg).parent

        with tempfile.NamedTemporaryFile(
            mode="w+",
            suffix=".toml",
            prefix="drytoml.",
            dir=str(parent),
        ) as fp:
            fp.write(document.as_string())
            fp.seek(0)
            yield fp


class Env(Wrapper):
    """Call another script, configuring it with an environment variable."""

    def __init__(self, env: Union[str, List[str]]):
        """Instantiate a cli wrapper.

        Args:
            env: Name(s) of the env var(s) to use which selects a
                 configuration file.
        """
        self.envs = (
            [
                env,
            ]
            if isinstance(env, str)
            else env
        )
        self.cfg = os.environ.get(self.envs[0], "pyproject.toml")

    def pre_import(self):
        """Configure env var before callback import."""
        for env in self.envs:
            os.environ[env] = self.virtual.name


class Cli(Wrapper):
    """Call another script, configuring it with specific cli flag."""

    def __init__(self, configs: List[str]):
        """Instantiate a cli wrapper.

        Args:
            configs: Possible names for the configuration flag of the
                wrapped script.

        Raises:
            ValueError: Empty configs.
        """
        if not configs:
            raise ValueError("No configuration strings received")

        for option in configs:
            try:
                idx = sys.argv.index(option)
                pre = sys.argv[:idx]
                post = sys.argv[idx + 2 :]
                cfg = sys.argv[idx + 1]
                break
            except ValueError:
                pass
        else:
            pre = sys.argv
            post = []
            cfg = "pyproject.toml"
            option = configs[0]
        self.cfg = cfg
        self.pre = pre
        self.post = post
        self.option = option

    def pre_call(self) -> None:
        """Prepare sys.argv to contain the configuration flag and file."""
        sys.argv = [*self.pre, self.option, f"{self.virtual.name}", *self.post]


def black():
    """Execute black, configured with custom setting cli flag."""
    Cli(["--config"])("black:patched_main")


def isort():
    """Execute isort, configured with custom setting cli flag."""
    Cli(["--sp", "--settings-path", "--settings-file", "--settings"])(
        "isort.main:main"
    )


def pylint():
    """Execute pylint, configured with custom setting cli flag."""
    Cli(["--rcfile"])("pylint:run_pylint")


def flakehell():
    """Execute flakehell, configured with custom env var."""
    Env(["FLAKEHELL_TOML", "PYLINTRC"])("flakehell:entrypoint")


def flake8helled():
    """Execute flake8helled, configured with custom env var."""
    Env(["FLAKEHELL_TOML", "PYLINTRC"])("flakehell:flake8_entrypoint")


def check():
    """Execute all formatters and linters, sequentially."""

    for command in (
        "dry -q isort .",
        "dry -q black .",
        "dry -q flakehell lint .",
    ):
        sp.run(shlex.split(command))  # noqa: S603, W1510
